/*
 *  Database record class for DPAP sharing
 *
 *  Copyright (C) 2008 W. Michael Petullo <mike@flyn.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <wand/wand_api.h>
#include <libexif/exif-data.h>
#include <libexif/exif-content.h>
#include <libexif/exif-entry.h>

#include "dmapd-dpap-record.h"

/* Media ID's start at max and go down.
 * Container ID's start at 1 and go up.
 */
static gint nextid = G_MAXINT;

struct DmapdDPAPRecordPrivate {
	gint id;
	gint filesize;
	gint largefilesize;
	const char *location;
	const char *filename;
	size_t thumbnail_size;
	unsigned char *thumbnail;
	const char *aspectratio;
	gint height;
	gint width;
	const char *format;
	const char *comments;
};

gint dmapd_dpap_record_get_id (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->id;
}

void dmapd_dpap_record_set_location    (DPAPRecord *record, const char *location)
{
	g_free (DMAPD_DPAP_RECORD (record)->priv->location);
	DMAPD_DPAP_RECORD (record)->priv->location = location;
}

const char *dmapd_dpap_record_get_location    (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->location;
}

const char *dmapd_dpap_record_get_title       (DPAPRecord *record)
{
	return "FIXME";
}

const char *dmapd_dpap_record_set_aspectratio (DPAPRecord *record, const char *aspectratio)
{
	g_free (DMAPD_DPAP_RECORD (record)->priv->aspectratio);
	DMAPD_DPAP_RECORD (record)->priv->aspectratio = aspectratio;
}

const char *dmapd_dpap_record_get_aspectratio (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->aspectratio;
}

void dmapd_dpap_record_set_filename (DPAPRecord *record, const char *filename)
{
	g_free (DMAPD_DPAP_RECORD (record)->priv->filename);
	DMAPD_DPAP_RECORD (record)->priv->filename = filename;
}

const char *dmapd_dpap_record_get_filename (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->filename;
}

const char *dmapd_dpap_record_get_format (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->format;
}

const char *dmapd_dpap_record_set_format (DPAPRecord *record, const char *format)
{
	g_free (DMAPD_DPAP_RECORD (record)->priv->format);
	DMAPD_DPAP_RECORD (record)->priv->format = format;
}

const char *dmapd_dpap_record_set_comments (DPAPRecord *record, const char *comments)
{
	g_free (DMAPD_DPAP_RECORD (record)->priv->comments);
	DMAPD_DPAP_RECORD (record)->priv->comments = comments;
}

const char *dmapd_dpap_record_get_comments (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->comments;
}

void dmapd_dpap_record_set_filesize (DPAPRecord *record, gint filesize)
{
	DMAPD_DPAP_RECORD (record)->priv->filesize = filesize;
}

gint        dmapd_dpap_record_get_filesize (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->filesize;
}

void dmapd_dpap_record_set_largefilesize (DPAPRecord *record, gint largefilesize)
{
	DMAPD_DPAP_RECORD (record)->priv->largefilesize = largefilesize;
}

gint        dmapd_dpap_record_get_largefilesize (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->largefilesize;
}

gint        dmapd_dpap_record_set_pixelheight (DPAPRecord *record, gint height)
{
	DMAPD_DPAP_RECORD (record)->priv->height = height;
}

gint        dmapd_dpap_record_get_pixelheight (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->height;
}

gint        dmapd_dpap_record_set_pixelwidth (DPAPRecord *record, gint width)
{
	DMAPD_DPAP_RECORD (record)->priv->width = width;
}

gint        dmapd_dpap_record_get_pixelwidth (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->width;
}

gint        dmapd_dpap_record_get_rating (DPAPRecord *record)
{
	return 5;
}

gint        dmapd_dpap_record_get_creationdate (DPAPRecord *record)
{
	return 1;
}

void        dmapd_dpap_record_set_thumbnail (DPAPRecord *record, unsigned char *thumb, size_t size)
{
	g_free (DMAPD_DPAP_RECORD (record)->priv->thumbnail);
	DMAPD_DPAP_RECORD (record)->priv->thumbnail = thumb;
	DMAPD_DPAP_RECORD (record)->priv->thumbnail_size = size;
}

unsigned char *dmapd_dpap_record_get_thumbnail (DPAPRecord *record, size_t *size)
{
	*size = DMAPD_DPAP_RECORD (record)->priv->thumbnail_size;
	return DMAPD_DPAP_RECORD (record)->priv->thumbnail;
}

GByteArray *dmapd_dpap_record_to_blob (DPAPRecord *record, size_t *size)
{
	/* FIXME: not endian safe. Don't write with one endianness and 
	 * read with another
	 */

	GByteArray *blob = g_byte_array_new ();

	*size = 0;

	/* NOTE: do not store ID in the blob. */

	g_byte_array_append (blob,
			     &(DMAPD_DPAP_RECORD (record)->priv->filesize),
			     sizeof (DMAPD_DPAP_RECORD (record)->priv->filesize));
	*size += sizeof (DMAPD_DPAP_RECORD (record)->priv->filesize);

	g_byte_array_append (blob,
			     &(DMAPD_DPAP_RECORD (record)->priv->largefilesize),
			     sizeof (DMAPD_DPAP_RECORD (record)->priv->largefilesize));
	*size += sizeof (DMAPD_DPAP_RECORD (record)->priv->largefilesize);

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->location,
			     strlen (DMAPD_DPAP_RECORD (record)->priv->location) + 1);
	*size += strlen (DMAPD_DPAP_RECORD (record)->priv->location) + 1;

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->filename,
			     strlen (DMAPD_DPAP_RECORD (record)->priv->filename) + 1);
	*size += strlen (DMAPD_DPAP_RECORD (record)->priv->filename) + 1;

	g_byte_array_append (blob,
			     &(DMAPD_DPAP_RECORD (record)->priv->thumbnail_size),
			     sizeof (DMAPD_DPAP_RECORD (record)->priv->thumbnail_size));
	*size += sizeof (DMAPD_DPAP_RECORD (record)->priv->thumbnail_size);

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->thumbnail,
			     DMAPD_DPAP_RECORD (record)->priv->thumbnail_size);
	*size += DMAPD_DPAP_RECORD (record)->priv->thumbnail_size;

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->aspectratio,
			     strlen (DMAPD_DPAP_RECORD (record)->priv->aspectratio) + 1);
	*size += strlen (DMAPD_DPAP_RECORD (record)->priv->aspectratio) + 1;

	g_byte_array_append (blob,
			     &(DMAPD_DPAP_RECORD (record)->priv->height),
			     sizeof (DMAPD_DPAP_RECORD (record)->priv->height));
	*size += sizeof (DMAPD_DPAP_RECORD (record)->priv->height);

	g_byte_array_append (blob,
			     &(DMAPD_DPAP_RECORD (record)->priv->width),
			     sizeof (DMAPD_DPAP_RECORD (record)->priv->width));
	*size += sizeof (DMAPD_DPAP_RECORD (record)->priv->width);

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->format,
			     strlen (DMAPD_DPAP_RECORD (record)->priv->format) + 1);
	*size += strlen (DMAPD_DPAP_RECORD (record)->priv->format) + 1;

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->comments,
			     strlen (DMAPD_DPAP_RECORD (record)->priv->comments) + 1);
	*size += strlen (DMAPD_DPAP_RECORD (record)->priv->comments) + 1;

	return blob;
}

static void dmapd_dpap_record_init (DmapdDPAPRecord *record)
{
	record->priv = DMAPD_DPAP_RECORD_GET_PRIVATE (record);
	record->priv->id = nextid--;
}

static void dmapd_dpap_record_class_init (DmapdDPAPRecordClass *klass)
{
	g_type_class_add_private (klass, sizeof (DmapdDPAPRecordPrivate));
}

static void dmapd_dpap_record_interface_init (gpointer iface, gpointer data)
{
	DPAPRecordInterface *dpap_record = iface;

	g_assert (G_TYPE_FROM_INTERFACE (dpap_record) == TYPE_DPAP_RECORD);

	dpap_record->get_id = dmapd_dpap_record_get_id;
	dpap_record->get_location = dmapd_dpap_record_get_location;
	dpap_record->get_title = dmapd_dpap_record_get_title;
	dpap_record->get_aspectratio = dmapd_dpap_record_get_aspectratio;
	dpap_record->get_filename = dmapd_dpap_record_get_filename;
	dpap_record->get_creationdate = dmapd_dpap_record_get_creationdate;
	dpap_record->get_format = dmapd_dpap_record_get_format;
	dpap_record->get_filesize = dmapd_dpap_record_get_filesize;
	dpap_record->get_largefilesize = dmapd_dpap_record_get_largefilesize;
	dpap_record->get_pixelheight = dmapd_dpap_record_get_pixelheight;
	dpap_record->get_pixelwidth = dmapd_dpap_record_get_pixelwidth;
	dpap_record->get_rating = dmapd_dpap_record_get_rating;
	dpap_record->get_comments = dmapd_dpap_record_get_comments;
	dpap_record->get_thumbnail = dmapd_dpap_record_get_thumbnail;
}

G_DEFINE_TYPE_WITH_CODE (DmapdDPAPRecord, dmapd_dpap_record, G_TYPE_OBJECT, 
			 G_IMPLEMENT_INTERFACE (TYPE_DPAP_RECORD, dmapd_dpap_record_interface_init))

DmapdDPAPRecord *dmapd_dpap_record_new_from_blob (GByteArray *blob)
{
	DmapdDPAPRecord *record = DMAPD_DPAP_RECORD (g_object_new (TYPE_DMAPD_DPAP_RECORD, NULL));

	guint8 *ptr = blob->data;

	/* NOTE: ID not in blob. */

	record->priv->filesize = *((gint *) ptr);
	ptr += sizeof (record->priv->filesize);

	record->priv->largefilesize = *((gint *) ptr);
	ptr += sizeof (record->priv->largefilesize);

	record->priv->location = g_strdup ((char *) ptr);
	ptr += strlen ((char *) ptr) + 1;

	record->priv->filename = g_strdup ((char *) ptr);
	ptr += strlen ((char *) ptr) + 1;

	record->priv->thumbnail_size = *((gint *) ptr);
	ptr += sizeof (record->priv->thumbnail_size);

	record->priv->thumbnail = g_memdup (ptr, record->priv->thumbnail_size);
	ptr += record->priv->thumbnail_size;

	record->priv->aspectratio = g_strdup ((char *) ptr);
	ptr += strlen ((char *) ptr) + 1;

	record->priv->height = *((gint *) ptr);
	ptr += sizeof (record->priv->height);

	record->priv->width = *((gint *) ptr);
	ptr += sizeof (record->priv->width);

	record->priv->format = g_strdup ((char *) ptr);
	ptr += strlen ((char *) ptr) + 1;

	record->priv->comments = g_strdup ((char *) ptr);
	ptr += strlen ((char *) ptr) + 1;

	return record;
}

DmapdDPAPRecord *dmapd_dpap_record_new (const char *path)
{
	MagickWand *wand;
	DmapdDPAPRecord *record = NULL;
	wand = NewMagickWand ();

	g_warning ("Calling MagickReadImage...");
	if (! MagickReadImage (wand, path)) {
		g_warning ("could not read %s", path);
	} else {
		ExifData *d;
		struct stat buf;
		char *format;
		guchar *thumb;
		size_t size;
		gint height;
		gint width;
		float aspectratio;

		g_warning ("MagickReadImage done");
		record = DMAPD_DPAP_RECORD (g_object_new (TYPE_DMAPD_DPAP_RECORD, NULL));

		if (stat (path, &buf) == -1) {
			g_warning ("Unable to determine size of %s", path);
		} else {
			dmapd_dpap_record_set_largefilesize (record, buf.st_size);
		}

		dmapd_dpap_record_set_location (record, g_strdup_printf ("file://%s", path));
		dmapd_dpap_record_set_filename (record, g_basename (path));

		MagickResetIterator (wand);
		format = MagickGetImageFormat (wand);
		dmapd_dpap_record_set_format (record, format);

		height = MagickGetImageHeight (wand);
		dmapd_dpap_record_set_pixelheight (record, height);

		width = MagickGetImageWidth (wand);
		dmapd_dpap_record_set_pixelwidth (record, width);

		aspectratio = width / (float) height;
		dmapd_dpap_record_set_aspectratio (record, g_strdup_printf ("%f", aspectratio));

		g_warning ("Calculating thumbnail...");
		if (width > height) {
			MagickResizeImage (wand,
					   106,
					   (1 / aspectratio) * 106,
					   LanczosFilter,
					   1.0);
		} else {
			MagickResizeImage (wand,
					   aspectratio * 106,
					   106,
					   LanczosFilter,
					   1.0);
		}
		thumb = MagickWriteImageBlob (wand, &size);
		dmapd_dpap_record_set_thumbnail (record, thumb, size);
		dmapd_dpap_record_set_filesize (record, size);

		dmapd_dpap_record_set_comments (record, g_strdup (""));
		/* FIXME:
		d = exif_data_new_from_file (path);
		if (! d) {
			g_warning ("Failed to EXIF data from %s", path);
		} else {
			int i;
			for (i = 0; i < EXIF_IFD_COUNT; i++) {
				ExifContent *c = d->ifd[i];
				if (! c || ! c->count) {
					g_warning ("Failed to find EXIF content in %s", path);
				} else {
					ExifEntry *e = exif_content_get_entry (c, EXIF_TAG_USER_COMMENT);
					if (! e) {
						g_warning ("Failed to get comments EXIF entry in %s", path);
					} else {
						gchar v[BUFSIZ + 1];
						exif_entry_get_value (e, v, BUFSIZ);
						dmapd_dpap_record_set_comments (record, g_strdup (v));
					exif_content_unref (c);
					}
				}
			exif_data_unref (d);
			}
		}
		*/
	}

	DestroyMagickWand (wand); 

	return record;
}
