/*   FILE: dmapd.c -- A DMAP server
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 01 January 2009 
 *
 * Copyright (c) 2009 W. Michael Petullo <new@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib/gprintf.h>
#include <glib/gdir.h>
#include <glib/gfileutils.h>
#include <libdmapsharing/dmap-mdns-publisher.h>
#include <libdmapsharing/dpap-share.h>

#include "dmapd-dmap-container-record.h"
#include "dmapd-dmap-container-db.h"
#include "dmapd-dpap-record.h"
#include "dmapd-dpap-db.h"

static gint
add_blob_to_db (const GByteArray *blob, DPAPDb *db)
{
	gint id = 0;
	DPAPRecord *record = dmapd_dpap_record_new_from_blob (blob);

	if (! record) {
		g_warning ("Could not load blob");
	} else {
		dmapd_dpap_db_add (db, record);
		id = dmapd_dpap_record_get_id (record);
	}

	return id;
}

static gint
add_file_to_db (const char *path, DPAPDb *db)
{
	gint id = 0;
	DPAPRecord *record = dmapd_dpap_record_new (g_strdup_printf ("%s", path));
	if (! record) {
		g_warning ("Could not load %s", path);
	} else {
		dmapd_dpap_db_add (db, record);
		id = dmapd_dpap_record_get_id (record);
	}

	return id;
}

static
gchar *cache_path (gchar *imagepath)
{
	gchar *cachepath = NULL;
	guchar hash[33] = {0};

	dmap_hash_generate (1, (const guchar*) imagepath, 2, hash, 0);
	cachepath = g_strdup_printf ("/var/cache/dmapd/%s", hash);

	return cachepath;
}

static
GByteArray *cache_read (gchar *imagepath)
{
	gchar *data;
	size_t size;
	GByteArray *blob = NULL;
	gchar *cachepath;
	GError *error = NULL;

	cachepath = cache_path (imagepath);

	g_file_get_contents (cachepath, &data, &size, &error);
	if (error != NULL) {
		g_warning ("No record cached at %s", cachepath);
	} else {
		blob = g_byte_array_new ();
		g_byte_array_append (blob, data, size);
		g_free (data);
	}

	g_free (cachepath);

	return blob;
}

static 
void cache_store (gchar *imagepath, GByteArray *blob, size_t size)
{
	gchar *cachepath;
	GError *error = NULL;
	cachepath = cache_path (imagepath);
	g_file_set_contents (cachepath, blob->data, size, &error);
	if (error != NULL) {
		g_error ("Error writing %s", cachepath);
	}

	g_free (cachepath);
}

static
void build_db_starting_at (const char *dir, DPAPDb *db, DMAPContainerDb *container_db, DMAPContainerRecord *container_record)
{
	GError *error = NULL;

	GDir *d = g_dir_open (dir, 0, &error);

	if (error != NULL) {
		g_error ("Error opening %s", dir);
	} else {
		const gchar *entry;
		
		while (entry = g_dir_read_name (d)) {
			gchar *path = g_strdup_printf ("%s/%s", dir, entry);

			if (g_file_test (path, G_FILE_TEST_IS_DIR)) {
				g_warning ("Search %s", path);
				DMAPContainerRecord *record = dmapd_dmap_container_record_new (g_strdup (entry));
				build_db_starting_at (path, db, container_db, record);
				dmapd_dmap_container_db_add (container_db, record);
			} else {
				GByteArray *blob;
				gint id;
				g_warning ("Add %s", path);
				blob = cache_read (path);
				if (blob) {
					id = add_blob_to_db (blob, db);
				} else {
					size_t size;
					DmapdDPAPRecord *record;

					id = add_file_to_db (path, db);
					record = dmapd_dpap_db_lookup_by_id (db, id);
					blob = dmapd_dpap_record_to_blob (record, &size);
					cache_store (path, blob, size);
				}
				if (id) {
					DPAPRecord *record = dmapd_dpap_db_lookup_by_id (db, id);
					if (container_record) {
						DMAPContainerEntry *entry = g_new (DMAPContainerEntry, 1);
						entry->media_id = id;
						entry->container_id = dmapd_dmap_container_record_get_id (container_record);
						dmapd_dmap_container_record_add_entry (container_record, entry);
					}
				}
				g_byte_array_free (blob, TRUE);
			}
			g_free (path);
		}
		
		g_dir_close (d);
	}
}

static char *
default_share_name ()
{
        const gchar *real_name;

        real_name = g_get_real_name ();
        if (strcmp (real_name, "Unknown") == 0) {
                real_name = g_get_user_name ();
        }

        return g_strdup_printf ("%s's Media", real_name);

	g_free (real_name);
}

static void
create_share (DPAPDb *db, DMAPContainerDb *container_db)
{
	char *name;

	DPAPShare *share = NULL;

	gpointer entry_type = NULL;

	g_warning ("initialize DAAP sharing");

	name = default_share_name ();
	share = dpap_share_new (name, NULL, db, container_db, entry_type);

	g_free (name);
}

static void
daemonize (void)
{
	int child, fd;
	char *pid;

	if (getppid () == 1)
		return; /* Already a daemon */

	child = fork ();

	if (child < 0)
		g_error ("Error forking");
	
	if (child > 0)
		exit (EXIT_SUCCESS);
	
	setsid ();

	chdir (RUNDIR);

	fd = open (LOCKPATH, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

	if (lockf (fd, F_TLOCK, 0) < 0)
		g_error ("Error opening lock file");
	
	fd = open (RUNDIR "/dmapd.pid", O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

	pid = g_strdup_printf ("%d", getpid ());
	write (fd, pid, strlen (pid));
	g_free (pid);
}

int main (int argc, char *argv[])
{
	char *root;
	DPAPDb *db;
	DMAPContainerDb *container_db;
	static GMainLoop *loop;

	if (argc != 2)
		g_fprintf (stderr, "usage: %s root\n", argv[0]);

	daemonize ();

	g_type_init ();
	g_thread_init (NULL);

	db = DPAP_DB (dmapd_dpap_db_new ());
	container_db = DMAP_CONTAINER_DB (dmapd_dmap_container_db_new ());
	g_chdir (argv[1]);
	build_db_starting_at (argv[1], db, container_db, NULL);

	loop = g_main_loop_new (NULL, FALSE);

	create_share (DPAP_DB (db), DMAP_CONTAINER_DB (container_db));

	g_main_loop_run (loop);

	exit(EXIT_SUCCESS);
}
